;; because of (array-data) and so...
(define (vm-dump-code cb . inner)

  (define (shorten-object o)
    (cond ((closure? o)
	   (format "#<closure ~a>" (shorten-object (closure-src o))))
	  ((identifier? o)
	   (format "#<identifier ~s#~s>" (id-name o)
		   (let ((lib (id-library o)))
		     (cond ((and lib (library? lib))
			    (library-name lib))
			   ((not lib) '#:toplevel)
			   (else lib)))))
	  ((library? o)
	   (format "#<library ~s>" (library-name o)))
	  ((pair? o)
	   (if (and (pair? (car o))
		    (not (zero? (length o))))
	       (let loop ((r '())
			  (o o))
		 (if (null? o)
		     (reverse r)
		     (loop (cons (shorten-object (car o)) r)
			   (cdr o))))
	       (let ((s (x->string o)))
		 (if (> (string-length s) 30)
		     (substring s 0 30)
		     s))))
	  ((box? o)
	   (format "#<box ~a>"
		   (shorten-object (unbox o))))
	  ((procedure? o)
	   (format "#<subr ~s>" (procedure-name o)))
	  ((macro? o)
	   (format "#<macro ~s>" (macro-name o)))
	  ((vector? o)
	   (format "#<code? ~a>" (shorten-object (vector-ref o 0))))
	  (else
	   o)))
  (define label?
    (lambda (l)
      (and (vector? l)
	   (> (vector-length l) 0)
	   (eqv? (vector-ref l 0) 11 #;$LABEL))))

  (define (unwrap sexp)
    (let loop ((sexp sexp)
	       (r '()))
      (cond ((null? sexp)
	     (reverse r))
	    ((identifier? sexp)
	     (id-name sexp))
	    ((pair? sexp)
	     (loop (car sexp)
		   (loop (cdr sexp)
			 r)))
	    (else
	     sexp))))
  
  (let ((data (array-data (code-builder-code cb)))
	(length (array-length (code-builder-code cb))))
    (format #t "~asize: ~a~%" (if (null? inner) "" (car inner)) length)
    ;(format #t "original code: ~%~s~%~%" data)
    (when (null? inner)
      (print "VM code:"))
    (let loop ((i 0))
      (if (>= i length)
	  (newline)
	  (let ((insn (vector-ref data i)))
	    (cond ((label? insn)
		   (format #t "#<label>~%")
		   (loop (+ i 1)))
		  (else
		   (let ((info (lookup-insn-name (get-insn insn))))
		     (flush-output-port (current-output-port))
		     (or (pair? info)
			 (errorf "invalid insn: ~a" insn))
		     (let ((name (car info))
			   (inst-values (cadr info))
			   (argc (caddr info))
			   (src? (cadddr info)))
		       (format #t "~a    ~a ~s" (if (null? inner) "" (car inner)) i name)
		       (let inst ((i 0))
			 (unless (= i inst-values)
			   (if (= i 0)
			       (display "(")
			       (display " "))
			   (display (get-insn-value insn inst-values i))
			   (inst (+ i 1))))
		       (unless (zero? inst-values)
			 (display ")"))
		       (let loop2 ((j 1))
			 (if (> j argc)
			     (begin
			       (if src?
				   (begin
				     (display " ;;")
				     (cond ((assv i (code-builder-src cb))
					    => (lambda (src)
						 (format #t " ~,,,,50s" (unwrap (cdr src))))))))
			       ;(format #t "  ;; ~x" insn)
			       (newline)
			       (flush-output-port (current-output-port))
			       (loop (+ i j)))
			     (let ((arg (vector-ref data (+ i j))))
			       (cond  ((code-builder? arg)
				       (begin
					 (format #t " ~a" (vector-ref arg 0))
					 (newline)
					 (format #t "~a Closure: ~%" (if (null? inner) "" (car inner)))
					 (vm-dump-code arg (string-append "  " (if (null? inner) ""
										   (car inner))))))
				      ((label? arg)
				       (format #t " #<label> "))
				      (else
				       (format #t " ~a" (shorten-object arg))))
			       (loop2 (+ j 1))))))))))))))
;;;; end of file
;; Local Variables:
;; coding: utf-8-unix
;; End:
